from . import versy
import argparse
import sys
import traceback

ACTIONS = versy.ACTIONS + ('<new version number>',)


def main(argv=None):
    args = parse(argv)
    try:
        versy.versy(**args)
    except Exception as e:
        print('ERROR:', e, file=sys.stderr)
        if args['verbose']:
            traceback.print_exc()
        sys.exit(-1)


def parse(argv=None):
    p = argparse.ArgumentParser(description=_DESCRIPTION)

    p.add_argument('action', nargs='?', default='show', help=_ACTIONS_HELP)
    p.add_argument('--changelog', '-c', default=None, help=_CHANGELOG_HELP)
    p.add_argument('--dry-run', '-d', action='store_true', help=_DRY_RUN_HELP)
    p.add_argument('--edit', '-e', action='store_true', help=_EDIT_HELP)
    p.add_argument('--file', '-f', default=None, help=_FILE_HELP)
    p.add_argument('--message', '-m', help=_MESSAGE_HELP)
    p.add_argument('--push', '-p', action='store_true', help=_PUSH_HELP)
    p.add_argument('--root', '-r', default='.', help=_ROOT_HELP)
    p.add_argument('--verbose', '-v', action='store_true', help=_VERBOSE_HELP)

    return vars(p.parse_args(argv))


def run(**kwargs):
    versy.versy(**dict(parse([]), kwargs))


_DESCRIPTION = 'Update the version number and change log of the program'
_ACTIONS_HELP = 'Actions are: ' + ', '.join(ACTIONS)
_CHANGELOG_HELP = 'Specify the name of a new change log'
_EDIT_HELP = 'Bring up the change log in an editor'
_FILE_HELP = 'Specify the file containing __version__'
_MESSAGE_HELP = 'Set change log message'
_ROOT_HELP = 'Root directory to search for a version number'
_PUSH_HELP = 'git push after committing changes'
_VERBOSE_HELP = 'Print more stuff'
_DRY_RUN_HELP = 'Don\'t actually make the changes, just print the diffs'


if __name__ == '__main__':
    sys.exit(main())



from pathlib import Path
import wolk

PREFIX = '__version__ = '
VERSION = 'VERSION'


class VersionFile:
    def __init__(self, root, printer, file=None):
        self.printer = printer
        if file:
            self.file = Path(file)
            self.version = _get_version(self.file)
        else:
            results = []
            for f in wolk.python(root):
                f = Path(f)
                version = _get_version(f)
                if version:
                    results.append((f, version))

            if not results:
                raise ValueError('No version files found')

            if len(results) > 1:
                res = ', '.join('{1} in {0}'.format(*r) for r in results)
                raise ValueError('More than one version file found: ' + res)

            self.file, self.version = results[0]

    def write(self, version):
        with self.printer(self.file) as _print:
            if self.file.name == VERSION:
                _print(version)

            else:
                for line in self.file.read_text().splitlines():
                    if line.startswith(PREFIX):
                        line = "%s'%s'" % (PREFIX, version)
                    _print(line)


def _get_version(f):
    if f.name == VERSION:
        return f.read_text().strip()
    if f.suffix != '.py':
        return

    lines = [s for s in f.read_text().splitlines() if s.startswith(PREFIX)]

    if not lines:
        return

    if len(lines) > 1:
        raise ValueError('More than one line starting "%s"' % PREFIX)

    line = lines[0]
    line = line[len(PREFIX) :]
    line = line.split('#')[0].strip()

    for quote in '"', "'":
        if len(line) > 2 and line[0] == line[-1] == quote:
            line = line[1:-1]

    return line



from unittest import TestCase
from versy.changelog import ChangeLog
import safer
import tdir


class TestChangeLog(TestCase):
    def setUp(self):
        self.results = []

    def test_simple(self):
        with tdir('setup.py', 'README') as td:
            cl = ChangeLog(td, '0.2.1', None, safer.printer, 'one', False)
            cl.new()

            actual = (td / 'CHANGELOG').read_text().splitlines()
            expected = ['## v0.2.1', '', '* one']
            assert expected[0].startswith(expected[0])
            assert actual[1:] == expected[1:]

            cl = ChangeLog(td, '0.2.1', None, safer.printer, 'two', False)

            with self.assertRaises(ValueError) as m:
                cl.new()
            assert m.exception.args[0].endswith('already exists')

            cl.update('0.3.0')
            actual = (td / 'CHANGELOG').read_text().splitlines()
            expected = [
                '## v0.3.0',
                '',
                '* two',
                '',
                '## v0.2.1',
                '',
                '* one',
            ]
            assert len(expected) == len(actual)
            for e, a in zip(expected, actual):
                if e.startswith('##'):
                    assert a.startswith(e)
                else:
                    assert a == e



from . import git
from . import semver
from .changelog import ChangeLog
from .version_file import VersionFile
import contextlib
import editor
import functools
import io
import myers
import safer

PREFIX = '__version__ = '
VERSION = 'VERSION'
ACTIONS = 'patch', 'minor', 'major', 'new', 'show'
__version__ = '0.9.6'


def versy(
    action, changelog, dry_run, file, message, root, verbose, edit, push
):
    printer = _dry_printer if dry_run else _printer

    if not dry_run:
        git.check_clean_workspace()

    vfile = VersionFile(root, printer, file)
    version = semver.semver(vfile.version)

    if action == 'show':
        print('Version', version, 'found in', vfile.file)
        return

    vb = verbose and not dry_run
    cl = ChangeLog(root, version, changelog, printer, message, vb)
    files = [cl.changelog]

    if action == 'new':
        cl.new()
        print('First version', version, 'to', cl.changelog)
        new_version = str(version)

    else:
        new_version = str(semver.bump(version, action))
        cl.update(new_version)
        vfile.write(new_version)
        if dry_run or verbose:
            print()
        print('Version', version, '-->', new_version, 'to', cl.changelog)

        if dry_run or verbose:
            print()
        files.append(vfile.file)

    if edit and not dry_run:
        editor(filename=cl.changelog)

    msg = 'Version v%s' % new_version
    git.commit(files, msg, dry_run, verbose, push)


@contextlib.contextmanager
def _dry_printer(file):
    fp = io.StringIO()

    yield functools.partial(print, file=fp)

    if file.exists():
        before = file.read_text().splitlines()
    else:
        before = []
    after = fp.getvalue().splitlines()

    print()
    print(60 * '-')
    print()
    print('%s:' % file)
    print()
    print(*myers.diff(before, after, context=2, format=True), sep='\n')


@contextlib.contextmanager
def _printer(file):
    with safer.printer(file) as _print:
        yield _print



import semver as _semver


def semver(s):
    try:
        return _semver.VersionInfo.parse(s)
    except Exception:
        raise ValueError('"%s" was not a valid version number' % s)


def bump(version, action):
    if action == 'patch':
        return version.bump_patch()
    if action == 'minor':
        return version.bump_minor()
    if action == 'major':
        return version.bump_major()
    try:
        v = _semver(action)
    except ValueError:
        raise ValueError('Unknown action %s' % action)

    if v <= version:
        raise ValueError('%s <= current version %s' % (v, version))

    return v



from . import git
from datetime import date
from pathlib import Path

NAMES = 'CHANGELOG', 'CHANGELIST', 'CHANGES', 'changelog'
SUFFIXES = {'', '.rst', '.txt', '.md'}


class ChangeLog:
    def __init__(self, root, version, changelog, printer, message, verbose):
        self.root = Path(root)
        self.version = version
        self.printer = printer
        self.changelog = changelog and Path(changelog)
        self.message = message
        self.verbose = verbose

        if not self.changelog:
            files = []
            for name in NAMES:
                for f in self.root.iterdir():
                    if f.stem == name and f.suffix in SUFFIXES:
                        files.append(f)
            if len(files) > 1:
                raise ValueError('Multiple changelogs: ' + ' '.join(files))
            if files:
                self.changelog = files[0]

    def log(self, *args, **kwds):
        if self.verbose:
            print(*args, **kwds)

    def log_print(self, *args, **kwds):
        self._print(*args, **kwds)
        self.log(' ', *args, **kwds)

    def new(self):
        changelog = Path(self.changelog or self.root / NAMES[0])
        if changelog.exists():
            raise ValueError('File %s already exists' % changelog.absolute())

        self.log(f'Creating {changelog}:')
        with self.printer(changelog) as self._print:
            self._entry(self.version, [])
            self.log_print('* %s' % (self.message or 'First release'))

        self.changelog = changelog

    def update(self, new_version):
        if not self.changelog:
            msg = 'Couldn\'t find a CHANGE file in %s' % self.root
            raise FileNotFoundError(msg)

        if self.message:
            messages = [self.message]
        else:
            messages = git.get_commits(str(self.version), self.root)
            messages = messages or [f'New version {self.version}']

        printed = False
        needs_empty_line = False
        with self.printer(self.changelog) as self._print:
            print(f'Writing {self.changelog}')
            if self.verbose:
                print()
            for line in self.changelog.read_text().splitlines():
                if not printed and str(self.version) in line:
                    printed = True
                    if needs_empty_line:
                        self._print()
                    self._entry(new_version, messages)
                    if messages:
                        self._print()
                self._print(line)
                needs_empty_line = bool(line.strip())

            if not printed:
                if needs_empty_line:
                    self._print()
                self._entry(new_version, messages)

    def _entry(self, version, messages):
        today = date.today().strftime('%y/%m/%d')
        title = 'v%s - %s' % (version, today)
        if self.changelog and self.changelog.suffix == '.rst':
            self.log_print(title)
            self.log_print('=' * len(title))
        else:
            self.log_print('##', title)
        self.log_print()
        for message in messages:
            self.log_print('*', message)



import subprocess

SUB = {'stderr': subprocess.DEVNULL, 'encoding': 'utf8'}
CMD = 'git log --pretty=format:%s HEAD~{end}..HEAD~{begin}'


def get_commits(version, cwd, max_commits=256, commit_block=2):
    commits = []

    begin = 0
    end = commit_block
    while True:
        cmd = CMD.format(begin=begin, end=end)
        try:
            out = subprocess.check_output(cmd.split(), cwd=cwd, **SUB)
        except subprocess.CalledProcessError:
            return []

        for line in out.splitlines():
            if line:
                if version in line:
                    return commits
                if len(commits) >= max_commits:
                    return []
                commits.append(line)

        begin += commit_block
        end += commit_block

    return commits


def check_clean_workspace():
    try:
        subprocess.check_output('git diff-index --quiet HEAD --'.split())
    except subprocess.CalledProcessError:
        raise ValueError('Cannot create new version with changes in workspace')


def commit(files, message, dry_run, verbose, push=False):
    cmds = [
        ('git', 'add') + tuple(str(f) for f in files),
        ('git', 'commit', '-m', message),
    ]
    if push:
        cmds.append(('git', 'push'))

    for cmd in cmds:
        if verbose or dry_run:
            print('$', *cmd)
        if not dry_run:
            if verbose:
                subprocess.run(cmd).check_returncode()
            else:
                subprocess.check_output(cmd)


if __name__ == '__main__':
    import sys

    for line in get_commits(sys.argv[1], '.'):
        print(line)



from unittest import TestCase
from versy import version_file
import tdir


class TestVersionFile(TestCase):
    def test_simple(self):
        with tdir(
            'setup.py',
            'README',
            'change.txt',
            sub={'bar': 'one', 'VERSION': '0.1.2'},
        ) as td:
            vf = version_file.VersionFile(td, None)
            assert vf.version == '0.1.2'
            assert vf.file.name == 'VERSION'

    def test_in_python(self):
        with tdir(
            'setup.py',
            'README',
            'change.txt',
            sub={'bar': 'one', 'foo.py': '__version__ = "0.1.2"'},
        ) as td:
            vf = version_file.VersionFile(td, None)
            assert vf.version == '0.1.2'
            assert vf.file.name == 'foo.py'



_classifiers = [
    'Development Status :: 4 - Beta',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
    'Topic :: Software Development :: Libraries',
    'Topic :: Utilities',
]


def _version():
    with open('versy/versy.py') as fp:
        line = next(i for i in fp if i.startswith('__version__'))
        return line.strip().split()[-1].strip("'")


with open('requirements.txt') as f:
    REQUIRED = f.read().splitlines()

if __name__ == '__main__':
    from setuptools import setup

    setup(
        name='versy',
        version=_version(),
        author='Tom Ritchford',
        author_email='tom@swirly.com',
        url='https://github.com/rec/versy',
        tests_require=['pytest'],
        py_modules=['versy'],
        description='Update the Python version number',
        long_description=open('README.rst').read(),
        license='MIT',
        classifiers=_classifiers,
        keywords=['documentation'],
        scripts=['scripts/versy'],
        packages=['versy'],
        install_requires=REQUIRED,
    )
